#!perl -T
use 5.006;
use strict;
use warnings;
use Test::More;

unless ( $ENV{RELEASE_TESTING} ) {
    plan( skip_all => "Author tests not required for installation" );
}

# Ensure a recent version of Test::Pod::Coverage
my $min_tpc = 1.08;
eval "use Test::Pod::Coverage $min_tpc";
plan skip_all => "Test::Pod::Coverage $min_tpc required for testing POD coverage"
    if $@;

# Test::Pod::Coverage doesn't require a minimum Pod::Coverage version,
# but older versions don't recognize some common documentation styles
my $min_pc = 0.18;
eval "use Pod::Coverage $min_pc";
plan skip_all => "Pod::Coverage $min_pc required for testing POD coverage"
    if $@;
    

#JWT:WE WANT TO SKIP POD COVERAGE ON "PRIVATE" MODULES LIKE "StreamFinder::_Class"!:
my @submodules = all_modules();
my $testplan = 0;
#1: CALCULATE # MODULES WE WANT TO TEST:
foreach my $submodule (@submodules) {
	$testplan++  unless ($submodule =~ /\:\:\_/o);
}

#TEST ALL NON-PRIVATE MODULES:
plan tests => $testplan;
foreach my $submodule (@submodules) {
	pod_coverage_ok($submodule)  unless ($submodule =~ /\:\:\_/o);
}

#all_pod_coverage_ok();  #JWT:SEEMS 2B NO WAY TO EXCLUDE SPECIFIC MODULES?!
